/*-----------------------------------------------------------------------
 *
 * File Name: Comm.h
 *
 * Author: Allen, Bruce ballen@dirac.phys.uwm.edu
 *
 * Revision: $Id: Window.h,v 1.1 1999/10/05 01:38:00 jolien Exp $
 *
 *-----------------------------------------------------------------------
 *
 * NAME
 * Window.h
 *
 * SYNOPSIS
 * #include "Comm.h"
 *
 * DESCRIPTION
 * Error codes, typedefs and prototypes required for use of the
 * Communication functions for binary inspiral searching
 *
 * DIAGNOSTICS
 *
 *-----------------------------------------------------------------------
 *
 * REVISION HISTORY
 *
 *
 *
 *-----------------------------------------------------------------------
 */

#ifndef _COMM_H
#define _COMM_H

#ifndef _LALDATATYPES_H
#include "LALDatatypes.h"
#define _LALDATATYPES_H
#endif

#ifndef _AVFACTORIES_H
#include "AVFactories.h"
#define _AVFACTORIES_H
#endif

#ifndef _MPI_INCLUDE
#include "mpi.h"
#define _MPI_INCLUDE
#endif

static CHAR *COMMH = "$Id:$";

#define COMM_EXPORTFAILED 1
#define COMM_MPIERROR 2
#define COMM_NULLPARA 4
#define COMM_UNRECMESG 8
#define COMM_NEGATIVE 16
#define COMM_NULLBUFF 32
#define COMM_NULLENV 64
#define COMM_NOENV 128
#define COMM_PUTENV 128
#define COMM_NULLSTRING 256
#define COMM_SUBPROC 512


#define COMM_MSGEXPORTFAILED "Failed to properly export environment variables"
#define COMM_MSGMPIERROR     "MPI Communications Error"
#define COMM_MSGNULLPARA     "NULL Parameter Block"
#define COMM_MSGUNRECMESG    "Unrecognized Message"
#define COMM_MSGNEGATIVE     "Number of items to transmit < 0"
#define COMM_MSGNULLBUFF     "Pointer to object buffer is NULL!"
#define COMM_MSGNULLENV      "Environment variable name is NULL"
#define COMM_MSGNOENV        "Environment variable not set"
#define COMM_MSGPUTENV       "The putenv() system call failed"
#define COMM_MSGNULLSTRING   "A text string in the parameter block was null"
#define COMM_MSGSUBPROC      "A subprocess did not appear to start correctly"

/* Structure for identifying processors */
typedef struct
{
    INT4 numprocs;  /* The total number of processes (N)*/
    INT4 myid;      /* The id of this process 0,...,N-1 */
    INT4 namelen;   /* The length in characters of this process's name */
    CHAR processor_name[MPI_MAX_PROCESSOR_NAME]; /* process name */
} mpiid;

/*
 * FUNCTIONS/STRUCTURES FOR DESCRIBING MESSAGES
 */

/* define the different types of messages to exchange  */
typedef enum {Finished,Request,Templates,Data,Events,Error} messtype;
#define NUMBERMESSAGES ((INT4)(Error+1))
#define MESSAGETYPELIST {"Finished","Request","Templates",\
                         "Data","Events","Error"}

/* input/output structure for ExchangeMessage() function */
typedef struct tagMessageExchange
{
    INT4 send;             /* set nonzero to send, zero to receive */
    messtype message;      /* message to send or message received */
    INT4 source;           /* used as output if send==0: where did message come from? */
    CHAR *type;            /* text specifying type of message received */
} MessageExchange;

void ExchangeMessage(Status* status,MessageExchange *params);

/* structure to describe a single template */
typedef struct
{
    REAL4 m1;            /* mass # 1 (in solar masses) */   
    REAL4 m2;            /* mass # 2 (in solar masses) */  
    INT4 templatenumber; /* the (unique identifying) template number */
    REAL4 fitfact;       /* the fitting factor with which bank was defined */
} tempstruct;

/*
 * FUNCTIONS/STRUCTURES FOR REQUESTING TEMPLATES
 */

/* structure to pass a template request */
typedef struct tagTemplateRequest
{
    INT4 initialize;       /* set to unity if initial bank is desired */
    INT4 templatenumber;   /* template number about which grid is needed */
    REAL4 fitfact;         /* desired fitting factor */
    tempstruct template;   /* the template about which you wish grid */
} TemplateRequest;

typedef struct tagRequestExchange
{
    INT4 send;               /* set nonzero to send, zero to receive */
    INT4 source;             /* used as input if send==0: slave to get template request from */
    TemplateRequest list;    /* the template request */
} RequestExchange;

void ExchangeRequest(Status* status,RequestExchange *params);


/*
 * FUNCTIONS/STRUCTURES FOR TEMPLATE EXCHANGE
 */

/* input/output structure for ExchangeTemplates() */
typedef struct tagTemplateExchange
{
    INT4 send;              /* set nonzero to send, zero to receive */
    INT4 destination;       /* used as input if send==1: slave to send templates to */
    INT4 number;            /* the number of templates being sent/received */
    tempstruct *list;       /* the list of templates being sent/received */
} TemplateExchange;

/* function for exchanging template lists */
void ExchangeTemplates(Status* status,TemplateExchange* params);

/*
 * FUNCTIONS/STRUCTURES FOR EXCHANGING DATA
 */

/* structure to pass a packet of data */
#define SGMTSZ 512
static const INT4 sgmtsz=SGMTSZ;
typedef struct tagDataPacket
{
    INT4 sgmt;              /* segment number */
    REAL8 time;             /* segment start time */
    REAL8 rate;             /* sampling rate in Hz */
    REAL4 data[SGMTSZ];     /* DMRO data */
    REAL4 resp[SGMTSZ+2];   /* response information */
    REAL4 wght[SGMTSZ/2+1]; /* weighting factor (inverse spectrum) */
} DataPacket;

#undef SGMTSZ

typedef struct tagDataExchange
{
    INT4 send;              /* set nonzero to send, zero to receive */
    INT4 destination;       /* used as input if send==1: slave to send data to */
    INT4 number;            /* the number of data packets to send */
    DataPacket *list;       /* the list of data packets */
} DataExchange;

void ExchangeData(Status* status,DataExchange *params);

/*
 * FUNCTIONS/STRUCTURES FOR EXCHANGING EVENTS
 */

/* structure to pass an event list */
typedef struct tagEventPacket
{
    INT4 sgmt;             /* segment number */
    INT4 tmpl;             /* template number */
    REAL8 time;            /* event time */
    REAL4 snr;             /* event SNR value */
    REAL4 chisq;           /* event chi-squared value */
    REAL4 sigma;           /* sensitivity of template */
} EventPacket;

typedef struct tagEventExchange
{
    INT4 send;              /* set nonzero to send, zero to receive */
    INT4 source;            /* used as input if send==0: slave to get events from */
    INT4 number;            /* the number of event packets to send */
    EventPacket  *list;     /* the list of event packets */
} EventExchange;

void ExchangeEvents(Status* status,EventExchange *params);

/* For passing environment variables */
typedef struct tagExportParams
{
	mpiid identity; /* the identity of the caller */
	CHAR* name;     /* the name of the environment variable to export from master->slaves */
    CHAR* value;    /* if non-null, set to this value rather than master's value */
} ExportParams;

void ExportEnvironment(Status *status,ExportParams* params);

/* For debugging */
typedef struct tagDebugMPIParams
{
	CHAR* debugger;	    /* the name of the debugger to invoke */
	CHAR* programname;  /* the name of the program to debug */
    INT4 startup;        /* set nonzero if you want to start debugger on this node */
    INT4 delaytime;      /* startup delay time in seconds */
    mpiid identity;     /* what machine are we? */	
} DebugMPIParams;

void DebugMPI(Status* status,DebugMPIParams* params);
#endif





