#ifndef _MATH_H
#include <math.h>
#ifndef _MATH_H
#define _MATH_H
#endif
#endif

#ifndef _LALSTDLIB_H
#include "LALStdlib.h"
#ifndef _LALSTDLIB_H
#define _LALSTDLIB_H
#endif
#endif

#ifndef _SEQFACTORIES_H
#include "SeqFactories.h"
#ifndef _SEQFACTORIES_H
#define _SEQFACTORIES_H
#endif
#endif

#ifndef _SPECBUFFER_H
#include "SpecBuffer.h"
#ifndef _SPECBUFFER_H
#define _SPECBUFFER_H
#endif
#endif

#ifndef _DATABUFFER_H
#include "DataBuffer.h"
#ifndef _DATABUFFER_H
#define _DATABUFFER_H
#endif
#endif

static volatile const char *DATABUFFERC = "$Id: DataBuffer.c$";

static void
FillDataBlock (
    Status     *status,
    DataBuffer *buffer,
    INT4        whichBlock
    )
{
  FrameDataOut  frameout;
  FrameDataInp  frameinp;
  DataBlock    *theBlock;

  INITSTATUS (status, DATABUFFERC);
  ATTATCHSTATUSPTR (status);

  theBlock = buffer->blockArray + whichBlock;
  frameout.framedata = theBlock->framedata;

  /* get next block of locked data */
  theBlock->continuous = 1;
  theBlock->anomalous = 0;
  theBlock->number = buffer->blocksFilled++;
  frameout.newLock = 0;

  frameinp.seek = 0;
  frameinp.seekPoints = 0;
  TRY (GetFrameData (status->statusPtr, &frameout, &frameinp, buffer->frameDataPar), status);

  while (frameout.newLock)
  {
    theBlock->continuous = 0;
    theBlock->anomalous = 1;
    buffer->newLock = 1;       /* set flag to indicate a new lock is pending */
    frameinp.seek = 1;
    frameinp.seekPoints = 3*60/frameout.framedata->deltaT; /* 3 minutes */
    TRY (GetFrameData (status->statusPtr, &frameout, &frameinp, buffer->frameDataPar), status);
  }

  /* if the block is not anomalous, add it to the spectrum buffer */
  if (!theBlock->anomalous)
  {
    TRY (AddSpectrum (status->statusPtr, buffer->specBuffer, theBlock->framedata), status);
  }

  if (frameout.newCal) /* new calibration information is pending */
    buffer->newCal = 1;

  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}


static void
FillAllDataBlocks (
    Status     *status,
    DataBuffer *buffer
    )
{
  INT4 block;

  INITSTATUS (status, DATABUFFERC);
  ATTATCHSTATUSPTR (status);

  buffer->newLock      = 0; /* new lock no longer pending */

  for (block = 0; block < buffer->blockArraySize; ++block)
  {
    TRY (FillDataBlock (status, buffer, block), status);

    if (buffer->newLock) /* need to refill from beginning */
    {
      block = -1;
      buffer->newLock = 0;
    }
  }

  buffer->nextData     = 0;
  buffer->nextBlock    = 0;
  buffer->deltaPoints  = 0;
  buffer->newLock      = 0; /* new lock no longer pending */
  
  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}


void
CreateDataBuffer (
    Status         *status,
    DataBuffer    **buffer,
    DataBufferPar  *params
    )
{
  SpectrumBufferPar      specpar;
  CreateVectorSequenceIn vseqin;
  INT4 block;

  INITSTATUS (status, DATABUFFERC);
  ATTATCHSTATUSPTR (status);

  /* make sure that arguments are not NULL */
  ASSERT (buffer, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (params, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);

  /* make sure that the buffer has not already been assigned */
  ASSERT (*buffer == NULL, status, DATABUFFER_ENNUL, DATABUFFER_MSGENNUL);

  /* assign memory for buffer */
  *buffer = (DataBuffer *) LALMalloc (sizeof (DataBuffer));

  /* make sure that the allocation was successful */
  ASSERT (*buffer, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);

  (*buffer)->blockArraySize = params->numSpec;
  vseqin.length             = params->numSpec + 1; /* extra swap space */
  vseqin.vectorLength       = params->numPoints;

  /* create buffer */
  (*buffer)->dataBuffer = NULL;
  TRY (I2CreateVectorSequence (status->statusPtr, &(*buffer)->dataBuffer, &vseqin), status);

  /* create block array */
  (*buffer)->blockArray = (DataBlock *)
    LALMalloc ((*buffer)->blockArraySize*sizeof(DataBlock));
  ASSERT ((*buffer)->blockArray, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);

  for (block = 0; block < (*buffer)->blockArraySize; ++block)
  {
    DataBlock *thisBlock = (*buffer)->blockArray + block;

    /* allocate memory for the framedata */
    thisBlock->framedata = (INT2TimeSeries *)
      LALMalloc (sizeof (INT2TimeSeries));
    ASSERT (thisBlock->framedata, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
    thisBlock->framedata->data = (INT2Vector *)LALMalloc (sizeof(INT2Vector));
    ASSERT (thisBlock->framedata->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
    /* should create units vector ... */
    thisBlock->framedata->data->length = params->numPoints;
    thisBlock->framedata->data->data   = (*buffer)->dataBuffer->data +
      block*params->numPoints;
  }

  /* create spectrum buffer */
  (*buffer)->specBuffer = NULL;
  specpar.numSpec    = params->numSpec;
  specpar.numPoints  = params->numPoints;
  specpar.windowType = params->windowType;
  specpar.plan       = params->plan;
  TRY (CreateSpectrumBuffer (status->statusPtr, &(*buffer)->specBuffer, &specpar), status);

  /* initialize frame data */
  (*buffer)->frameDataPar = NULL;
  TRY (InitializeFrameData (status->statusPtr, &(*buffer)->frameDataPar), status);

  (*buffer)->blocksFilled = 0;
  (*buffer)->newCal       = 0;
  (*buffer)->numSent      = 0;
  (*buffer)->first        = 1;

  /* fill all data blocks */
  TRY (FillAllDataBlocks (status->statusPtr, *buffer), status);

  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}


void
DestroyDataBuffer (
    Status      *status,
    DataBuffer **buffer
    )
{
  INT4 block;

  INITSTATUS (status, DATABUFFERC);
  ATTATCHSTATUSPTR (status);

  /* make sure that arguments are not NULL */
  ASSERT (buffer, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (*buffer, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);

  /* finalize frame data */
  TRY (FinalizeFrameData (status->statusPtr, &(*buffer)->frameDataPar), status);

  /* destroy spectrum buffer */
  TRY (DestroySpectrumBuffer (status->statusPtr, &(*buffer)->specBuffer), status);

  /* destroy data buffer */
  TRY (I2DestroyVectorSequence (status->statusPtr, &(*buffer)->dataBuffer), status);

  for (block = 0; block < (*buffer)->blockArraySize; ++block)
  {
    DataBlock *thisBlock = (*buffer)->blockArray + block;
    LALFree (thisBlock->framedata->data);
    LALFree (thisBlock->framedata);
  }

  LALFree ((*buffer)->blockArray);
  LALFree (*buffer);
  *buffer = NULL;

  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}


void
GetData (
    Status     *status,
    GetDataOut *output,
    INT4        advance,
    DataBuffer *buffer
    )
{
  INT4 numPoints;
  INT4 numBlocks;

  INITSTATUS (status, DATABUFFERC);
  ATTATCHSTATUSPTR (status);

  /* make sure that arguments are not NULL */
  ASSERT (output, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (buffer, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->spec, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->resp, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->data->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->spec->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->resp->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->data->data->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->spec->data->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);
  ASSERT (output->resp->data->data, status, DATABUFFER_ENULL, DATABUFFER_MSGENULL);

  /* check sizes ... */
  numPoints = output->data->data->length;
  ASSERT (numPoints > 0, status, DATABUFFER_ESIZE, DATABUFFER_MSGESIZE);
  ASSERT (buffer->dataBuffer->vectorLength == numPoints, status, DATABUFFER_ESZMM, DATABUFFER_MSGESZMM);
  ASSERT (output->resp->data->length == numPoints, status, DATABUFFER_ESZMM, DATABUFFER_MSGESZMM);
  ASSERT (output->spec->data->length == numPoints/2 + 1, status, DATABUFFER_ESZMM, DATABUFFER_MSGESZMM);

  /* make sure advance is reasonable */
  ASSERT (advance > 0, status, DATABUFFER_ESIZE, DATABUFFER_MSGESIZE);
  ASSERT (advance < numPoints, status, DATABUFFER_ESIZE, DATABUFFER_MSGESIZE);

  numBlocks = buffer->blockArraySize;

  if (buffer->first)
  {
    output->newLock = 1;
    buffer->first   = 0;
  }
  else
  {
    output->newLock = 0;
  }
  output->newCal = 0;

  /* does next block contain continuous data?  (only need to check if
   * buffer->nextData is not at the beginning of a block)
   */
  if (buffer->nextData%numPoints > 0)
  {
    INT4       nextBlockNo = (buffer->nextData/numPoints + 1)%numBlocks;
    DataBlock *nextBlock   = buffer->blockArray + nextBlockNo;
    INT4       continuous  = nextBlock->continuous;

    if (!continuous) /* no: skip into next block */
    {
      output->newLock = 1;
      /* refill buffer */
      TRY (FillAllDataBlocks (status->statusPtr, buffer), status);
    }
  }

  /* do we need to fill another block of data? (only do so if a
   * buffer->newLock flag is not pending)
   */
  if (buffer->deltaPoints > numBlocks*numPoints/2 && !buffer->newLock)
  {
    TRY (FillDataBlock (status->statusPtr, buffer, buffer->nextBlock), status);

    buffer->deltaPoints -= numPoints;

    if (buffer->nextBlock == 0) /* copy data to end of buffer */
    {
      INT2 *first = buffer->dataBuffer->data;
      INT2 *last  = first + numBlocks*numPoints;
      memcpy (last, first, numPoints*sizeof(INT2));
    }

    buffer->nextBlock = (buffer->nextBlock + 1)%numBlocks;
  }

  /* copy the data to the output vector */
  memcpy (
      output->data->data->data,
      buffer->dataBuffer->data + buffer->nextData,
      numPoints*sizeof(INT2)
      );

  /* set epoch, etc ... */
  {
    INT4            blockno = buffer->nextData/numPoints;
    INT2TimeSeries *tseries = (buffer->blockArray + blockno)->framedata;
    REAL4           time;
    time  = tseries->epoch.gpsSecs + 1e-9*tseries->epoch.gpsNsecs;
    time += (buffer->nextData%numPoints)*tseries->deltaT;
    
    output->data->epoch.gpsSecs  = time;
    output->data->epoch.gpsNsecs = fmod (1e9*time, 1e9);
    output->data->deltaT = tseries->deltaT;
    output->data->f0     = tseries->f0;
  }

  /* get average spectrum */
  TRY (AverageSpectrum (status->statusPtr, output->spec, buffer->specBuffer), status);

  /* now is the time to deal with new calibration if pending */
  if (buffer->newCal && output->newLock)
  {
    output->newCal = 1;

    /* get new calibration data */
    TRY (CalibrateFrameData (status->statusPtr, output->resp, buffer->frameDataPar), status);

    buffer->newCal = 0;
  }

  output->number = buffer->numSent++;

  /* bookkeeping */
  buffer->deltaPoints += advance;
  buffer->nextData    += advance;

  /* if next data is in end of buffer, move it back to the beginning */
  if (buffer->nextData/numPoints == numBlocks)
    buffer->nextData = buffer->nextData%numPoints;

  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}
